#pragma once

#include "MinimalAllocator.hpp"
#include <atlpath.h>

namespace Minimal {

namespace MinimalString {

__inline
size_t MinimalStringLength(const wchar_t *str)
{
	return ::lstrlenW(str);
}

__inline
size_t MinimalStringLength(const char *str)
{
	return ::lstrlenA(str);
}

__inline
wchar_t * MinimalStringCopy(wchar_t *dst, const wchar_t *src)
{
	return ::lstrcpyW(dst, src);
}

__inline
char * MinimalStringCopy(char *dst, const char *src)
{
	return ::lstrcpyA(dst, src);
}

}

template<typename CharType>
class MinimalStringT
{
public:
	typedef CharType * StringType;
	typedef const CharType * ConstStringType;

public:
	explicit MinimalStringT(IMinimalAllocator *alloc) :
		m_alloc(alloc), 
		m_str((CharType *)alloc->Allocate(sizeof CharType)), m_msize(1), m_size(0)
	{
		m_str[0] = 0;
	}

	explicit MinimalStringT(ConstStringType str, IMinimalAllocator *alloc) :
		m_alloc(alloc)
	{
		m_size = MinimalString::MinimalStringLength(str);
		m_msize = m_size + 1;
		m_str = (CharType *)alloc->Allocate(m_msize * sizeof CharType);
		MinimalString::MinimalStringCopy(m_str, str);
	}

	~MinimalStringT() {
		m_alloc->Deallocate(m_str);
	}

public:
	void Grow(size_t N)
	{
		if(N <= m_size) return;
		void *p = m_alloc->Reallocate(m_str, N * sizeof CharType);
		m_str = (CharType *)p;
		m_msize = N;
	}

	void Repair()
	{
		m_size = MinimalString::MinimalStringLength(m_str);
	}

	StringType GetRaw()
	{
		return m_str;
	}

	size_t GetSize()
	{
		return m_size;
	}

	operator StringType()
	{
		return m_str;
	}

protected:
	IMinimalAllocator *m_alloc;
	StringType m_str;
	size_t m_msize;
	size_t m_size;
};

#ifdef MINIMAL_USE_PROCESSHEAPSTRING
__declspec(selectany)
ProcessHeapAllocator g_allocator;

template <typename CharType>
class ProcessHeapStringT : public MinimalStringT<CharType>
{
private:
	ProcessHeapStringT<CharType> & operator =(const ProcessHeapStringT<CharType> &str);

public:
	explicit ProcessHeapStringT()
		: MinimalStringT<CharType>(&g_allocator) {}

	explicit ProcessHeapStringT(const CharType *str)
		: MinimalStringT<CharType>(str, &g_allocator) {}

public:
	ProcessHeapStringT<CharType> & operator =(const CharType *str)
	{
		m_size = MinimalString::MinimalStringLength(str);
		Grow(m_size + 1);
		MinimalString::MinimalStringCopy(m_str, str);
		return *this;
	}

	ProcessHeapStringT<CharType> & operator +=(const CharType *str)
	{
		size_t strLen = MinimalString::MinimalStringLength(str);
		Grow(m_size + strLen + 1);
		MinimalString::MinimalStringCopy(m_str + m_size, str);
		m_size += strLen;
		return *this;
	}
};
#endif

}
